/*
 * fxtest
 * 
 * Test FXO/FXS interfaces with a delayed echo
 * 
 * FXO test:
 *   Call FXO interface. FXO will answer (offhook) and start audio echo test.
 * 
 * FXS test:
 *   Take phone offhook. FXS will start audio echo test.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>

#include <dahdi/user.h>
#include "../linux/include/dahdi/ap_user.h"

#define READ_SIZE	160
#define MAX_CHANNEL	8

static int dahdi_get_event(int fd)
{
	int x;
	if (ioctl(fd, DAHDI_GETEVENT, &x) == -1)
		return -1;
	return x;
}

static int dahdi_set_hook(int fd, int hook)
{
	return ioctl(fd, DAHDI_HOOK, &hook);
}

static int dahdi_echocan(int fd, int length)
{
	struct dahdi_echocanparams ecp = { .tap_length = length };
	return ioctl(fd, DAHDI_ECHOCANCEL_PARAMS, &ecp);
}

int open_channel(int chan)
{
	char devname[64] = "/dev/dahdi/channel";
	int fd;
	int x;
	
	/* Open DAHDI device channel */
	fd = open(devname, O_RDWR);
	if (fd < 0) {
		printf("Unable to open '%s': %s\n", devname, strerror(errno));
		return -1;
	}
	
	/* Specify DAHDI channel for FD */
	if (chan) {
		if (ioctl(fd, DAHDI_SPECIFY, &chan)) {
			x = errno;
			close(fd);
			errno = x;
			printf("Unable to specify channel %d: %s\n", chan, strerror(errno));
			return -1;
		}
	}
	
	/* Set block size to 160 samples or 20 ms */
	x = READ_SIZE;
	if (ioctl(fd, DAHDI_SET_BLOCKSIZE, &x) == -1) {
		printf("Unable to set blocksize '%d': %s\n", x,  strerror(errno));
		x = errno;
		close(fd);
		errno = x;
		return -1;
	}
	return fd;
}

void print_usage()
{
	fprintf(stderr, "Usage: fxtest [-n channel] [-h]\n");
	fprintf(stderr, "  -n\tChannel number (Default: 1)\n");
	fprintf(stderr, "  -t\Echo delay time in secs (Default: 1)\n");
	fprintf(stderr, "  -h\tThis help\n");
}

int running = 1;

static void signal_handler(int signal)
{
	running = 0;
}

int main(int argc, char *argv[])
{
	int opt;
	int fd;
	int chan = 1;
	unsigned char *buf;
	int offset = 0;
	int bytes;
	int offhook = 0;
	int buflen = 8000; /* default: 1 sec */
	
	/* Get parameters */
	while ((opt = getopt(argc, argv, "n:t:h")) != -1) {
		switch (opt) {
			case 'n':
				chan = atoi(optarg);
				break;
			case 't':
				buflen = atoi(optarg) * 8000;
				break;
			case 'h':
				print_usage();
				return 0;
				break;
			default: /* '?' */
				break;
		}
	}
	
	/* Handle signals to stop main loop*/
	signal(SIGHUP, signal_handler);
	signal(SIGINT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGALRM, signal_handler);
	
	/* Open DAHDI FX Channel */
	fd = open_channel(chan);
	if (fd < 0) {
		printf("Falha ao abrir canal %d\n", chan);
		return -1;
	}
	
	/* Starting test */ 
	printf("Testando canal %d\n", chan);
	
	buf = malloc(buflen);
	
	dahdi_set_hook(fd, DAHDI_ONHOOK);
	
	/* Main loop */
	while (running) {
		/* Send back old samples.
		 * Mute samples for a second to filter out hook clicking sound
		 */
		if (offhook && offhook++ < 50) {
			memset(buf, 0xd5, buflen);
		}
		write(fd, buf + offset, READ_SIZE);
		
		/* Read new samples */
		bytes = read(fd, buf + offset, READ_SIZE);
		if (bytes < 0) {
			/* Short read, should be an event */
			int event = dahdi_get_event(fd);
			
			switch (event) {
				case DAHDI_EVENT_ONHOOK:
					printf("Canal %d esta no gancho\n", chan);
					break;
				case DAHDI_EVENT_RINGBEGIN:
					printf("Canal %d esta chamando\n", chan);
					break;
				case DAHDI_EVENT_RINGOFFHOOK:
					printf("Canal %d esta fora do gancho. Realizar teste de audio.\n", chan);
					/* Open audio and enable echo cancellation */
					dahdi_set_hook(fd, DAHDI_OFFHOOK);
					dahdi_echocan(fd, 1024);
					offhook = 1;
					break;
				case -1:
					/* Ignore error */
					break;
				default:
					printf("DAHDI FX got event: %d/0x%08x\n", event, event);
					break;
			}
		}
		
		offset = (offset + READ_SIZE) % buflen;
	}
	
	/* Ending test */
	printf("Encerrando teste no canal %d\n", chan);
	dahdi_set_hook(fd, DAHDI_ONHOOK);
	dahdi_echocan(fd, 0);
	free(buf);
	close(fd);
	return 0;
}
